@@ -33,6 +33,7 @@
 #include "vino-input.h"
 #include "vino-cursor.h"
 #include "vino-prompt.h"
+#include "vino-status-icon.h"
 #include "vino-util.h"
 #include "vino-enums.h"
 #include <sys/poll.h>
@@ -54,6 +55,7 @@ struct _VinoServerPrivate
   VinoFB           *fb;
   VinoCursorData   *cursor_data;
   VinoPrompt       *prompt;
+  VinoStatusIcon   *icon;
 
   GIOChannel       *io_channel;
   guint             io_watch;
@@ -78,7 +80,7 @@ struct _VinoServerPrivate
   guint             use_alternative_port : 1;
 };
 
-typedef struct
+struct _VinoClient
 {
   rfbClientPtr  rfb_client;
   GIOChannel   *io_channel;
@@ -89,7 +91,7 @@ typedef struct
   guint         auth_timeout;
   char         *auth_response;
   int           auth_resp_len;
-} VinoServerClientInfo;
+};
 
 enum
 {
@@ -106,10 +108,10 @@ enum
   PROP_VNC_PASSWORD
 };
 
-static enum rfbNewClientAction vino_server_auth_client (VinoServer           *server,
-							VinoServerClientInfo *client,
-							const char           *response,
-							int                   length);
+static enum rfbNewClientAction vino_server_auth_client (VinoServer *server,
+							VinoClient *client,
+							const char *response,
+							int         length);
 
 static void vino_server_setup_framebuffer     (VinoServer *server);
 static void vino_server_release_framebuffer   (VinoServer *server);
@@ -117,12 +119,31 @@ static void vino_server_update_security_types (VinoServer *server);
 
 static gpointer parent_class;
 
+static void
+vino_server_client_accepted (VinoServer *server,
+                             VinoClient *client)
+{
+  if (!server->priv->icon)
+    server->priv->icon = vino_status_icon_new (server, server->priv->screen);
+
+  vino_status_icon_add_client (server->priv->icon, client);
+}
+
+static void
+vino_server_client_disconnected (VinoServer *server,
+                                 VinoClient *client)
+{
+  if (vino_status_icon_remove_client (server->priv->icon, client))
+    {
+      g_object_unref (server->priv->icon);
+      server->priv->icon = NULL;
+    }
+}
 static void
 vino_server_handle_client_gone (rfbClientPtr rfb_client)
 {
-  VinoServer           *server = VINO_SERVER (rfb_client->screen->screenData);
-  VinoServerClientInfo *client = NULL;
-  GSList               *l;
+  VinoServer *server = VINO_SERVER (rfb_client->screen->screenData);
+  GSList     *l;
 
   g_return_if_fail (VINO_IS_SERVER (server));
 
@@ -132,7 +153,7 @@ vino_server_handle_client_gone (rfbClientPtr rfb_client)
 
   for (l = server->priv->clients; l; l = l->next)
     {
-      client = (VinoServerClientInfo *) l->data;
+      VinoClient *client = (VinoClient *) l->data;
 
       if (rfb_client == client->rfb_client)
 	{
@@ -160,6 +181,9 @@ vino_server_handle_client_gone (rfbClientPtr rfb_client)
 
 	  g_free (l->data);
 	  server->priv->clients = g_slist_delete_link (server->priv->clients, l);
+
+          vino_server_client_disconnected (server, client);
+
 	  break;
 	}
     }
@@ -238,9 +262,9 @@ vino_server_client_data_pending (GIOChannel   *source,
 }
 
 static void
-vino_server_set_client_on_hold (VinoServer            *server,
-				VinoServerClientInfo  *client,
-				gboolean               on_hold)
+vino_server_set_client_on_hold (VinoServer  *server,
+				VinoClient  *client,
+				gboolean     on_hold)
 {
   rfbClientPtr rfb_client = client->rfb_client;
 
@@ -285,8 +309,8 @@ vino_server_set_client_on_hold (VinoServer            *server,
 static enum rfbNewClientAction
 vino_server_handle_new_client (rfbClientPtr rfb_client)
 {
-  VinoServer           *server = VINO_SERVER (rfb_client->screen->screenData);
-  VinoServerClientInfo *client;
+  VinoServer *server = VINO_SERVER (rfb_client->screen->screenData);
+  VinoClient *client;
 
   g_return_val_if_fail (VINO_IS_SERVER (server), RFB_CLIENT_REFUSE);
 
@@ -295,7 +319,7 @@ vino_server_handle_new_client (rfbClientPtr rfb_client)
   if (!server->priv->fb)
     vino_server_setup_framebuffer (server);
 
-  client = g_new0 (VinoServerClientInfo, 1);
+  client = g_new0 (VinoClient, 1);
 
   rfb_client->clientData = client;
 
@@ -317,13 +341,16 @@ vino_server_handle_prompt_response (VinoServer         *server,
 				    rfbClientPtr        rfb_client,
 				    VinoPromptResponse  response)
 {
+  VinoClient *client = (VinoClient *) rfb_client->clientData;
+
   g_return_if_fail (VINO_IS_SERVER (server));
   g_return_if_fail (rfb_client != NULL);
 
   switch (response)
     {
     case VINO_RESPONSE_ACCEPT:
-      vino_server_set_client_on_hold (server, rfb_client->clientData, FALSE);
+      vino_server_set_client_on_hold (server, client, FALSE);
+      vino_server_client_accepted (server, client);
       break;
     case VINO_RESPONSE_REJECT:
       rfbCloseClient (rfb_client);
@@ -338,13 +365,16 @@ vino_server_handle_prompt_response (VinoServer         *server,
 static enum rfbNewClientAction
 vino_server_handle_authenticated_client (rfbClientPtr rfb_client)
 {
-  VinoServer           *server = VINO_SERVER (rfb_client->screen->screenData);
-  VinoServerClientInfo *client = (VinoServerClientInfo *) rfb_client->clientData;
+  VinoServer *server = VINO_SERVER (rfb_client->screen->screenData);
+  VinoClient *client = (VinoClient *) rfb_client->clientData;
 
   g_return_val_if_fail (VINO_IS_SERVER (server), RFB_CLIENT_REFUSE);
 
   if (!server->priv->prompt_enabled)
-    return RFB_CLIENT_ACCEPT;
+    {
+      vino_server_client_accepted (server, client);
+      return RFB_CLIENT_ACCEPT;
+    }
 
   vino_prompt_add_client (server->priv->prompt, rfb_client);
 
@@ -412,7 +442,7 @@ vino_server_handle_clipboard_event (char         *str,
 }
 
 static gboolean
-vino_server_auth_client_deferred (VinoServerClientInfo *client)
+vino_server_auth_client_deferred (VinoClient *client)
 {
   VinoServer              *server = VINO_SERVER (client->rfb_client->screen->screenData);
   enum rfbNewClientAction  result;
@@ -436,10 +466,10 @@ vino_server_auth_client_deferred (VinoServerClientInfo *client)
 }
 
 static void
-vino_server_defer_client_auth (VinoServer           *server,
-			       VinoServerClientInfo *client,
-			       const char           *response,
-			       int                   length)
+vino_server_defer_client_auth (VinoServer *server,
+			       VinoClient *client,
+			       const char *response,
+			       int         length)
 {
   client->auth_resp_len = length;
   client->auth_response = g_new (char, length);
@@ -491,10 +521,10 @@ vino_server_get_password_from_keyring (VinoServer *server)
 }
 
 static enum rfbNewClientAction
-vino_server_auth_client (VinoServer           *server,
-			 VinoServerClientInfo *client,
-			 const char           *response,
-			 int                   length)
+vino_server_auth_client (VinoServer *server,
+			 VinoClient *client,
+			 const char *response,
+			 int         length)
 {
   rfbClientPtr  rfb_client;
   char         *password;
@@ -551,8 +581,8 @@ vino_server_check_vnc_password (rfbClientPtr  rfb_client,
 				const char   *response,
 				int           length)
 {
-  VinoServer           *server = VINO_SERVER (rfb_client->screen->screenData);
-  VinoServerClientInfo *client = (VinoServerClientInfo *) rfb_client->clientData;
+  VinoServer *server = VINO_SERVER (rfb_client->screen->screenData);
+  VinoClient *client = (VinoClient *) rfb_client->clientData;
 
   g_return_val_if_fail (VINO_IS_SERVER (server), FALSE);
 
@@ -693,7 +723,8 @@ static void
 vino_server_init_from_screen (VinoServer *server,
 			      GdkScreen  *screen)
 {
-  rfbScreenInfoPtr rfb_screen;
+  rfbScreenInfoPtr  rfb_screen;
+  char             *name;
 
   g_return_if_fail (server->priv->screen == NULL);
   g_return_if_fail (screen != NULL);
@@ -718,6 +749,15 @@ vino_server_init_from_screen (VinoServer *server,
 		  gdk_screen_get_height (screen),
 		  -1, -1, -1);
 
+  name = g_strjoin ("@",
+                    g_get_user_name (),
+                    g_get_host_name (),
+                    NULL);
+
+  rfbSetDesktopName(rfb_screen, name);
+
+  g_free (name);
+
   /* libvncserver NOTE:
    *   DeferUpdateTime is the number of milliseconds to wait
    *   before actually responding to a frame buffer update
@@ -1209,7 +1249,7 @@ vino_server_set_on_hold (VinoServer *server,
 
       for (l = server->priv->clients; l; l = l->next)
 	{
-	  VinoServerClientInfo *client = l->data;
+	  VinoClient *client = l->data;
 
 	  /* If a client is on hold before we have initialized,
 	   * we want to leave it on hold until the deferred
@@ -1354,3 +1394,16 @@ vino_server_get_vnc_password (VinoServer *server)
 
   return server->priv->vnc_password;
 }
+
+G_CONST_RETURN char *
+vino_client_get_hostname (VinoClient *client)
+{
+  return client->rfb_client->host;
+}
+
+void
+vino_client_disconnect (VinoClient *client)
+{
+  rfbCloseClient (client->rfb_client);
+  rfbClientConnectionGone (client->rfb_client);
+}
